module textop
   use model
! contains subroutines for
!       - reading and writing parameters
!       - for decomposing one line into alphanumeric and numeric 
!         components separated by spaces
!       - reading optional parameters from parameter files
! IM, 2/23/2000 - 12/15/2002
   integer :: max_string_readline = 800 
   CONTAINS


  subroutine read_parameters()
  ! reads and decodes parameter file
   implicit none
   integer n, x(40),ef,stat,curr_eff,i,j,k
   character xc(40)*40

   write(*,'(a)',advance='no')' name of parameter file?'
   read '(a)',parfile
   parfile=adjustl(parfile)               ! ignore leading spaces
   write(*,'(a)') parfile
   open(io_p,file=parfile)

   call chkfmtnewline(io_p)
 
   call readline(io_p,x,xc,n)
   if (n == -1) then
       print*,'INPUT FILE EMPTY'
       stop
   endif
   if (xc(1) /= 'DATAFILE') then
       print*,'DATAFILE not found'
       stop
   endif
   call readline(io_p,x,xc,n) 
   datafile=xc(1)

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'NUMBER_OF_TRAITS') then
       print*,'NUMBER_OF_TRAITS not found'
       stop
   endif
   call readline(io_p,x,xc,n) 
   ntrait=x(1)

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'NUMBER_OF_EFFECTS') then
       print*,'NUMBER_OF_EFFECTS not found'
       stop
   endif
   call readline(io_p,x,xc,n) 
   neff=x(1)

   allocate(pos_y(ntrait),pos_eff(neff,ntrait),&
            nlev(neff),effecttype(neff),nestedcov(neff,ntrait),&
            randomtype(neff),randomnumb(neff),randomfile(neff),&
            r(ntrait,ntrait),rinv(ntrait,ntrait),g(neff,maxcorr,maxcorr))
   randomtype=g_fixed
   randomnumb=0
   nestedcov=0
   g=0; r=0

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'OBSERVATION(S)') then
       print*,' OBSERVATION(S) not found'
       stop
   endif
   call readline(io_p,x,xc,n) 
   if (n == ntrait) then
        pos_y(1:ntrait)=x(1:ntrait)
       else
         print*,ntrait,' numbers expected after OBSERVATION(S)'
	 stop
   endif

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'WEIGHT(S)') then
       print*,' WEIGHT(S) not found'
       stop
   endif
   call readline(io_p,x,xc,n) 
   if (n == 1) then
        pos_weight=x(1)
       elseif (n == 0) then
        pos_weight=0
       else
         print*,ntrait,' Only one number for WEIGHT'
	 stop
   endif

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'EFFECTS:') then
       print*,' EFFECTS: not found'
       stop
   endif
   do ef=1,neff
      call readline(io_p,x,xc,n)
      if (n == ntrait+2 .or. n == 2*ntrait+2) then
          pos_eff(ef,:)=x(1:ntrait)
          nlev(ef)=x(ntrait+1)
          select case (xc(ntrait+2)(1:3))
               case ('cro')         !crossclassified
                  effecttype(ef)=effcross
               case ('cov')         !covariate
                  effecttype(ef)=effcov
                  if (n==2*ntrait+2) then
                       nestedcov(ef,:)=x(ntrait+3:n)
                  endif
               case default
                  print*,'Unknown type of effect: ',xc(ntrait+2)
          end select
        else
          print*,'too few or too many numbers for effect ',ef,'. Was ',n
          stop
      endif 
   enddo

   call readline(io_p,x,xc,n)
   if (xc(1) /= 'RANDOM_RESIDUAL') then
       print*,' RANDOM_RESIDUAL not found'
       stop
   endif
   read(io_p,*,iostat=stat)((r(i,j),i=1,ntrait),j=1,ntrait)
   if (stat.ne.0) then
      print*,'error reading RANDOM_RESIDUAL variances'
      stop
   endif


   do
       call readline(io_p,x,xc,n)
          if (n == -1) exit
          if (xc(1) /='RANDOM_GROUP') exit

          call readline(io_p,x,xc,n)
          if (n<1 .or. n>neff) then
             print*,'line after RANDOM_GROUP has too few or too many numbers'
             stop
          endif
          if (x(1)<1 .or. x(n)> neff) then
              print*,'number ', x(1:n),' after RANDOM_GROUP out of range'
              stop
          endif
          do i=1,n
             if (i+x(1)-1 /= x(i)) then
                print*,' correlated effects:',x(1:n),' should be consecutive'
                stop
             endif
          enddo
          curr_eff=x(1)
          randomnumb(curr_eff)=n
          call readline(io_p,x,xc,n)
          if (xc(1) /= 'RANDOM_TYPE') then
              print*,' RANDOM_TYPE not found'
              stop
          endif

          call readline(io_p,x,xc,n) 
          select case (xc(1)(1:11))
             case ('diagonal')
                randomtype(curr_eff)=g_diag
             case('add_animal')
                randomtype(curr_eff)=g_A
             case('add_an_upg')
                randomtype(curr_eff)=g_A_UPG
             case('add_an_upgi')
                randomtype(curr_eff)=g_A_UPG_INB
             case('add_sire')
                randomtype(curr_eff)=g_As
             case('par_domin')
                randomtype(curr_eff)=g_PD
	     case('user_file')
                randomtype(curr_eff)=g_usr
             case('user_file_i')
                randomtype(curr_eff)=g_usr_inv
             case('add_an_usym')
                randomtype(curr_eff)=g_A_UnSym
             case default
                print*,'unknown RANDOM-TYPE: ', xc(1)
                print*,' Should be diagonal, add_animal, add_sire',&
                       ' add_an_upg, add_an_upginb, add_an_usym '
		print*,'    user_file, user_file_inv, or par_domin'
                stop
          end select  

          call readline(io_p,x,xc,n)
          if (xc(1) /= 'FILE') then
              print*,' FILE for random effect ',curr_eff,' not found'
              stop
          endif
          call readline(io_p,x,xc,n)
          randomfile(curr_eff)=xc(1)
           !ANNE
       !   print *,'read_parameter  ',curr_eff,io_off,xc(1)
          if (randomtype(curr_eff) /= g_diag) open(curr_eff+io_off,file=xc(1))


          call readline(io_p,x,xc,n)
          if (xc(1) /= '(CO)VARIANCES') then
              print*,' (CO)VARIANCES not found'
              stop
          endif

          k=ntrait*randomnumb(curr_eff)
          if (k <= maxcorr) then
             read(io_p,*,iostat=stat)((g(curr_eff,i,j),i=1,k),j=1,k)
             if (stat /= 0) then
                print*,'error reading variances for effect ',curr_eff
                stop
             endif
           else
              print*,'maxcorr should be increased to at least ',k
              stop
          endif
   enddo
end subroutine


subroutine readline(unit,x,xc,n,nohash)
! Reads a line from unit and decomposes it into numeric fields in x and
! alphanumeric fields in xc; the number of fields is stored in n.
! If optional variable nohash is absent, 
!  all characters after character # are ignored.

  integer unit,x(:),n,stat,i
  integer,optional::nohash
  character xc(:)*(*)
  character(len=max_string_readline) :: a  
 
  do
     read(unit,'(a)',iostat=stat)a
     if (stat /= 0) then
         n=-1
         return
     endif
     a=adjustl(a)               ! ignore leading spaces
     
     if (present(nohash)) then 
        exit
       else	!ignore characters after #
        i=scan(a,'#')
        if (i == 0) exit
        if (i>1) then
            a=a(1:i-1)
            exit
        endif
     endif  	
  enddo
  
  call nums(a,x,xc,size(x),n)
  return
end subroutine


subroutine readline1(unit,x,xc,n)
! as readline but does not ignore entries after character #

  integer unit,x(:),n,stat,i,nohash
  character xc(:)*(*)  
 
  call readline(unit,x,xc,n,nohash)
end subroutine  



subroutine readlinef(unit,x,xc,n)
! As readline but decodes x to floating point

  integer unit,n,stat,i
  character (*)::xc(:)
  character(len=max_string_readline) :: a
  real::x(:)
 
  read(unit,'(a)',iostat=stat)a
  if (stat == 0) then
      a=adjustl(a)               ! ignore leading spaces
      call nums2(a,n,x,xc)
    else
      n=-1
  endif
end subroutine



subroutine print_parameters(title)
  ! Prints parameters of the model
   use denseop
   implicit none
   integer :: i,j,k,l
   character (20):: name
   character (40):: fmt
   character (*):: title
   logical::not_pd

   print*,'    ',title
 
   print '(/'' Parameter file:'',t30,a)',parfile
   print '('' Data file:'',t30,a)',datafile
   print '('' Number of Traits'',t30,i2)',ntrait
   print '('' Number of Effects'',t30,i2)',neff
   print '('' Position of Observations'',t30,20i3)',(pos_y(i),i=1,ntrait)
   print '('' Position of Weight (1)'',t30,20i3)',pos_weight
   print '('' Value of Missing Trait/Observation'',t40,i8)',miss

   print '(/,''EFFECTS'')'
   print '(a,t25,a,t45,a)',' #  type','position (2)','levels   [positions for nested]'
   do i=1,neff
      select case (effecttype(i))
           case (effcross)
              name='cross-classified'
           case (effcov)
              name='covariable'
           case default
              name='???'
      end select
 
      ! writing with number of fields dependent on ntrait and presence of cov
      write(fmt,'(''(i2,2x,a19,'',i2,''i3,t75,i10,t90,20i3)'')')ntrait
      if (nestedcov(i,1) /= 0) then
           print fmt,i,name,pos_eff(i,:),nlev(i),nestedcov(i,:)
          else
           print fmt,i,name,pos_eff(i,:),nlev(i)
      endif
   enddo

   print '(/'' Residual (co)variance Matrix'')'
   if (.not. checksym(r) ) then
      print*,'R Matrix not symmetric'
      stop 
   endif 
   call pos_def(r,'R  Matrix not positive definite',&
                1d-18,stat=not_pd)
   if (not_pd) stop	
   call printmat(r,fmt='(20g12.5)')
   
   
   do i=1,neff
      if (randomnumb(i) == 0) then
          cycle
      else if (randomnumb(i) == 1) then
          print '(/,'' Random Effect(s)'',t20,20i3)',(/(j,j=i,i+randomnumb(i)-1)/)
      else if(randomnumb(i) >1) then
           print '(/,'' correlated random effects'',t30,20i3)',&
                     (/(j,j=i,i-1+randomnumb(i)) /)
      endif
      select case (randomtype(i))
         case (g_diag)
            print '('' Type of Random Effect:'',t30,a)','diagonal'
         case(g_A)
            print '('' Type of Random Effect:'',t30,a)','additive animal'
            print '('' Pedigree File:'',t30,a)',randomfile(i)
         case(g_As)
            print '('' Type of Random Effect:'',t30,a)','additive sire'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
         case(g_A_UPG)
            print '('' Type of Random Effect:'',t30,a,a)','additive animal',&
                     ' with unknown parent groups'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
         case(g_A_UPG_INB)
            print '('' Type of Random Effect:'',t30,a,a)','additive animal',&
                      ' with unknown parent groups and inbreeding'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
         case(g_PD)
            print '('' Type of Random Effect:'',t30,a,a)','parental dominance'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
	 case(g_usr)
            print '('' Type of Random Effect:'',t30,a,a)',&
	            'user defined from file'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
	 case(g_usr_inv)
            print '('' Type of Random Effect:'',t30,a,a)',&
	           'user defined from file and inverted'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
         case(g_A_UnSym)
            print '('' Type of Random Effect:'',t30,a,a)','additive animal unsymmetric'
            print '('' Pedigree File: '',t30,a)',randomfile(i)
         case default
            print*,'unknown RANDOM-TYPE'
            stop
      end select  
      
      j=ntrait*randomnumb(i)   

      if (.not. checksym(g(i,:j,:j) )) then 
         print*,'G Matrix not symmetric'
         stop
      endif       
      call pos_def(g(i,:j,:j),'G  Matrix not positive definite',&
		   1d-18,stat=not_pd)
      if (not_pd) stop
     	   
      print '('' trait   effect    (CO)VARIANCES'')'
      
      do k=0,randomnumb(i)-1
         do j=1,ntrait
             l=j+k*ntrait
             print '(i3,i8,2x,50g12.4)',j,k+i,g(i,l,1:randomnumb(i)*ntrait)
         enddo
      enddo
   enddo
   

   print '(/'' REMARKS'')'
   print*,' (1) Weight position 0 means no weights utilized'
   print*,' (2) Effect positions of 0 for some effects and traits means that such'
   print*,'     effects are missing for specified traits'
   print*
end subroutine



subroutine nums2(a,n,x,xc)
! separates array a into items delimited by blanks. character elements are
! put into optional character vector xc, decoded numeric values 
! into optional real vector x, and n contains the number of items. The 
! dimension of x and xc can be lower than n.
! A modification of nums() from f77 to f90
! Now accepts real numbers
! 2/23/2000

character (*)::a
character (*),optional::xc(:)
real,optional::x(:)
integer::n,curr,first,last,lena,stat,i

curr=1;lena=len(a);n=0

do 
  ! search for first nonspace
  first=0
  do i=curr,lena
     if (a(i:i) /= ' ') then
        first=i
	exit
     endif
  enddo
  if (first == 0) exit
  
  
  ! search for first space
  curr=first+1
  last=0
  do i=curr,lena
      if (a(i:i) == ' ') then
        last=i
        exit
      endif
  enddo
  
  if (last == 0) last=lena
  
  n=n+1
  if (present(xc)) then
     if (size(xc) >= n) then
           xc(n)=a(first:last)
         else
              print*, "Error in nums2 splitting string starting: "
              print*,a(1:80)
              print*,'into m:',size(xc),'items'
           stop
     endif
  endif   
  if (present(x)) then
     if (size(x) >= n) then
            read(a(first:last),'(f12.0)',iostat=stat)x(n)
            if (stat /=0) x(n)=0
         else
              print*, "Error in nums2 splitting string starting: "
              print*,a(1:80)
              print*,'into m:',size(xc),'items'            
              stop
     endif
  endif   
	
  curr=last+1
enddo
end subroutine

subroutine getoption(name,n,x,xc)
! In unit io_p=40, which reads parameter line in BLUPF90, locates line:
! OPTION name str1 str2
! where str1, str2 are strings separated by spaces.
! Then, it assigns: xc(1)=str1,  xc(2)=str2,...
! and attempts to decode strings into real values: x1=value(str1),....
!
! n contains the number of strings.  x and xc are optional and their 
! dimensions may be smaller than n in which case some strings/values are
! not stored.
!
! Upon exit, unit io_p=40 points to line next to the one located.
!
! If the line cannot be located, n=-1
! 

character (*)::name
integer::n
real,optional::x(:)
character (*),optional::xc(:)

real::x1(1000)
integer::stat,m
character (50)::xc1(1000)
character (400)::a

rewind io_p
n=-1

do
   read (io_p,'(a)',iostat=stat)a
   if (stat /= 0) exit
      call nums2(a,m,x1,xc1)       
      if ( m>1 .and. xc1(1) == 'OPTION' .and. xc1(2) == name) then
         n=m-2
	 if (present(xc)) xc=xc1(3:size(xc)+2)
	 if (present(x)) x=x1(3:size(x)+2)
	 exit
      endif
enddo
end subroutine

subroutine getoption_r8(name,n,x,xc)
! As getoption() but with x in r8 precision
! 
character (*)::name
integer::n
real (r8)::x(:)
real::xr4(size(x))
character (*),optional::xc(:)
!
call getoption(name,n,xr4,xc)
x=xr4
end subroutine

subroutine chkfmtnewline(un)
character(50000) :: a,i_name ! big number to allow check genotypes files
integer :: i,un
!
inquire(un,name=i_name)
read(un,'(a)') a 
do i=1,len(a) 
   if (ichar(a(i:i)) == 13) then ! Mac files
      print '(/a,a,a)', ' Mac old newline format detected in file: "',trim(i_name(index(i_name,'/',back=.true.)+1:)),'"'
      print '(a)'  , ' Convert it to Unix newline format, e.g. "flip -u <file>"'
      stop
   endif
enddo
rewind(un)
endsubroutine

end module

