module cmdarg
! set of tools to process command arguments 
! Ignacio Aguilar
! iaguilar@inia.org.uy
! 
! Last modification: 4/21/2012
integer, parameter :: strlen=250
character(strlen),allocatable :: myarg(:)
contains 

subroutine printcmdarg(a,m,x,xc)
! print arguments
character (*) :: a
character (*),optional :: xc(:)
integer :: i,m
character (strlen) :: str
real,optional :: x(:)

write(*,'(a,a)',advance='no') '    ',trim(adjustl(a))
do i=1,m
write(*,'(a,a)',advance='no') ' ',trim(adjustl(xc(i))) 
enddo 
write(*,'(a)') '' 
 
end subroutine

subroutine getcmdarg(a,m,x,xc)
! read command arguments and return optional paramaters if present
! n = -1 if nor in list
! n >=0 number of values of arguments
character (*) :: a
character (*),optional :: xc(:)
integer :: i,j,n,m,ncmd,io
character (strlen) :: str
real,optional :: x(:)
!
if (present(xc)) xc=' '
if (present(x)) x=0

ncmd=0
n=0
m=-1
ncmd=command_argument_count()
do i=1,ncmd
   call get_command_argument(i,str)
   if (adjustl(str)==adjustl(a)) then
      n=n+1
      exit
   endif
enddo
do j=i+1, ncmd
   call get_command_argument(j,str)
   str=adjustl(str)
   if (str(1:2)=='--') exit
   n=n+1
   if (present(xc)) xc(n-1) = str
   if (present(x))  then
      read(xc(n-1),*,iostat=io) x(n-1)
      if (io/=0) x(n-1)=0
   endif
enddo
if (n>0)m=n-1
end subroutine

subroutine chkcmdarg(xa)
! check all arguments to be in the list of valid ones. 
character (*) :: xa(:)
integer :: i,n
character (strlen):: str
!
n=command_argument_count()
do i=1,n
   call get_command_argument(i,str)
   str=adjustl(str)
   if (str(1:1)=='-') then
   if (.not. any(xa==str) ) then
      print *,'Command argument NOT valid: ',trim(adjustl(str))
      stop
   endif
   endif
enddo
end subroutine

subroutine addcmdarg(a,xa)
! add argument to list of valid arguments for checks
character (*),allocatable :: xa(:)
character (*) :: a
character (strlen), allocatable :: tmp(:)
!
if (allocated(xa)) then
   n=size(xa)
   allocate(tmp(n))
   tmp=xa
   deallocate(xa)
   allocate(xa(n+1))
   xa(1:n)=tmp
   xa(n+1)=a
   deallocate(tmp)
else
   allocate(xa(1))
   xa(1)=a
endif
end subroutine

endmodule 

