module searchsubs

CONTAINS

!------------------
 subroutine sortmat (a,col,n,order,types)
!------------------
! I. Misztal, rewrite on an old subroutine
! 
! by default, sorts matrix of integers in ascending order column by column. 
! Optionally, only first n columns are sorted
! if optional order is present, order='d' sorts in descending order.
! if optional types is present, each integer may be also treated as
! alphanumeric or real depending on each character of types: i stands
! for integer, r for real and a for alphanumeric.
! For example, if col=(/4,5,2,3/) and types="aair", rows 4 and 5 will be 
! sorted alphanumerically, 2 as integer and 3 as real.
!

implicit none 
integer,optional::n 
character (*),optional::order,types
character::ch
integer :: i,j,s,l,r,j1,k,icol,ix
integer ::a(:,:) 
integer,dimension(size(a,dim=1))::x,y 
integer :: stack(50,2),col(:) 
integer,parameter::is_lt=1, is_eq=2, is_gt=3
logical::descending=.false.

if (present(order)) then
   if (order(1:1)/='a' .or. order(1:1) /='d') then
      print*,"sortmat: sort order is either 'a' for ascending or 'd' for ",&
            "descending"
      stop
   endif
   if (order(1:1)=='d') descending=.true.
 endif
 
 if (present(types)) then
    if (len(types)< size(col)) then     	    
        print*,"sortmat: ordering types of wrong length:'\",len(types),"<",&
	        size(col) 
	endif
	do i=1,size(col)
	   ch=types(i:i)
	   if (scan(ch,'air') == 0) then
	      print*,"sortmat: types should contain only characters 'iar'"
	      stop
	   endif
	enddo
 endif	       	

if (present(n)) then 
  ix=n 
else 
  ix=size(a,dim=2) 
endif
icol=size(col) 
s=1 
stack(1,1)=1; stack(1,2)=ix
do 
  l=stack(s,1); r=stack(s,2) 
  s=s-1
  do
    i=l; j=r 
    j1=(l+r)/2 !median element
    x(1:icol)=a(col,j1)
    do ! select larger than median starting from bottom 
      do
        k=compint(a(col,i),x(1:icol),descending,types)
        select case (k) 
          case (is_lt)
            i=i+1
            cycle
          case (is_eq,is_gt)
            exit
          case default
            print*,'should not happen 123548'
            stop
        end select
      enddo 
      !select smaller than median starting from top
      do
        k=compint(a(col,j),x(1:icol),descending,types)
        select case (k)
          case (is_gt)
            j=j-1
            cycle
          case (is_eq,is_lt)
            exit
          case default
            print*,'should not happen 123549' 
            stop
        end select
      enddo
    ! swap
    if (i.le.j) then
      y=a(:,i); a(:,i)=a(:,j); a(:,j)=y !swap 
      i=i+1; j=j-1
    endif
    if (i > j) exit
  enddo
    if (i.lt.r) then 
      s=s+1 
      if (s > 50) then 
        print*,'stack > 50' 
        stop 
      endif 
      stack(s,1)=i 
      stack(s,2)=r 
    endif
    r=j 
    if (l >= r) exit 
  enddo
  if (s == 0) exit 
enddo 
!----------------------
 end subroutine sortmat
!----------------------

!----------------
 function compint (x,y,descending,types) result (j)
!---------------- 
! compares integer vectors sequentially 
! j = 1 if x < y 
! 2 if x = y 
! 3 if x > y
! order is reversed if descending==true
!
! if optional types is present, integers can be treated also as alphanumeric 
! or real variables as documented before
!
implicit none 
integer::x(:),y(:),i, j,k
character (4)::ch
logical::descending,l
character (*),optional::types
real::r

if (.not.present(types)) then
   ! purely integer comparisons
   j=2 
   do i=1,size(x) 
     if (x(i) < y(i)) then 
      j=1 
      exit 
     elseif (x(i) > y(i)) then 
      j=3 
      exit 
     endif 
   enddo 
 else
   j=2 
   do i=1,size(x) 
     select case(types(i:i))
        case("i")	!integer
            if (x(i) < y(i)) then 
            j=1 
            exit 
          elseif (x(i) > y(i)) then 
            j=3 
            exit 
          endif 
	case('r')	!real
	  if (transfer(x(i),r) < transfer(y(i),r)) then 
            j=1 
            exit 
          elseif (transfer(x(i),r) > transfer(y(i),r)) then 
            j=3 
            exit 
          endif   
	case('a')	!alphanumeric
            ! this cause memory leak with Abosft 10.0.6
            ! if (transfer(x(i),ch) < transfer(y(i),ch)) then 
            l = transfer(x(i),ch) < transfer(y(i),ch)
            if (l) then
               j=1 
               exit 
            !elseif (transfer(x(i),ch) > transfer(y(i),ch)) then 
            else
               l = transfer(x(i),ch) > transfer(y(i),ch)
               if (l) then
                  j=3 
                  exit 
               endif   
           endif
   end select
 enddo
endif
 if (descending) j=4-j
!--------------------
 end function compint
!--------------------



subroutine tokenize(a,xc,n)
! Variable a contains tokens separated by spaces
! xc(1:n) returns those tokens and n contains the number of tokens
! This subroutine is slow
  implicit none
  
  character (*)::a,xc(:)
  integer::n,first,last
  
  first=1
  n=0
   
  do 
     first=verify(a(first:),' ')+first-1	     ! first nonspace
     last=scan(a(first+1:),' ')+first-1  ! last nonspace character
     
     if (last<first) then
         n=n+1
	 xc(n)=a(first:)
	 exit
      endif	 
    
     if (last >= first) then
           n=n+1
	   xc(n)=a(first:last)
	   if (first == last ) exit
	   first=last+1
         else
	   exit
      endif
   enddo
  
  end subroutine








end module
