! ============================================================================
! Name        : BuildMqqFromBlup.f03
! Author      : Olivier Filangi
! Version     :
! Copyright   : GPLv3
! Description :
! ============================================================================

program BuildMqqFromBlupf90
    use cli
    use omp_lib

    implicit none

    integer, parameter   :: dp = SELECTED_REAL_KIND(12, 60)
    real,parameter       :: zero = 0.0, half = 0.5, one = 1.0, two = 2.0,   &
                          vsmall = TINY(1.0), vlarge = HUGE(1.0)

    integer , parameter            :: MAXPOSITION = 100000
    character(len=len_long_cli)    :: nposstring="0",sizeid="10"
    character(len=1)               :: bshort=""

    ! La construction de Mqq va necessite de lire beacoup de fichier (50 000).
    ! Pour optimiser l'acces au disque sans saturer la memoire
    ! nous lisons les elements par paquets
    integer  , parameter :: sizeBuffSNP = 30

    !Arguments de l'application
    character(len=len_long_cli)   :: genotypeFile=""    ! fichier contenant les genotype au format AIPL
    character(len=len_long_cli)   :: patternMatFile=""  ! Pattern du nom des fichier contenant les information matriciel par calcul des positions SNP/Haplotypes
    integer              :: npos = 0           ! Nombre de position testé correspondant au nombre de SNP/position haplotypique testé
                                               ! permet de lire le nombre de fichier correspondant au pattern

    integer              :: nu = 0 ! number of animal genotyped

    integer                                        :: idread
    ! correspondance id à une renumerotation pour acceder aux structure de donnees
    integer           ,dimension(:) ,pointer       :: id       ! correspondance index animaux , numero identifiant
    integer           ,dimension(:) ,pointer       :: revid    ! correspondance identifiant animaux , index

    !status des animaux genotype
    integer , dimension(:,:)    , pointer          :: genotype
    character(len=MAXPOSITION)                     :: bufsnp

    integer                                        :: nbe ! number of fixed effect

    ! to read entries from blupf90
    integer            , dimension(:), allocatable :: idmax,ivec,jvec
    real(kind=dp)      , dimension(:), allocatable :: vvec

    ! to build matrix from entries Blupf90
    real(kind=dp)      , dimension(:,:), allocatable :: A
    real(kind=dp)      , dimension(:), allocatable   :: B
    real(kind=dp)      , dimension(:,:), allocatable :: C
    real(kind=dp)      , dimension(:,:), allocatable :: D
    real(kind=dp)      , dimension(:)  , allocatable :: Eq1,Eq2

    ! name of matrix file blupf90
    character(len=len_long_cli)                      :: filemat,animalFile,defstr

    logical            , dimension(:  ), allocatable :: matrix_loaded

    real(kind=dp)      , dimension(:,:), allocatable :: M,P,P2,P3,SIGMA_MAT
    real(kind=dp)      , dimension(:)  , allocatable :: SIGMA,P3diag
    real(kind=dp)      , dimension(:)  , allocatable :: U,V,Ts,Psmin,Pvalues ! Simulation structure
    double precision      , dimension(:,:), allocatable :: M2,P22,Q22
    double precision      , dimension(:), allocatable :: W,WORK

    integer :: ios,i,q1,q2,ndim,isim,s,ifail,def,is,js
    character(len=10) :: qstr
    character(len=1)  :: TAIL_G01ECF = 'U',JOBZ,UPLO
    real(kind=dp)     :: Fq1q2,vare,invvare,vtemp,SEUIL_SELECT_VAR=1.d-5

    integer :: nsim = 10,lwork,info,j,k
    logical :: sort
    character(len=len_long_cli) :: option_l
    character(len=1)            :: option_s_opt,JOBU,JOBVT

    !====================================================================================
    !
    !     Arguments
    !
    !====================================================================================
    write (*,*)
    write (*,*) "======================================================================"
    write (*,*) "--geno , -g     : genotype file"
    write (*,*) "--idlength, -i  : Length ID in the genotype file (10 ?)"
    write (*,*) "--matfile, -m   : pattern mat file"
    write (*,*) "--npos, -n      : number of position (SNP/Haplotype)"
    write (*,*) "--vare, -v      : residual variance"
    write (*,*) "--nlevel, -l    : number of level"
    write (*,*) "--animal, -a    : file with animal phenotyped"
    write (*,*) "--nsim, -s      : Number of simulation"
    write (*,*)
    write (*,*)

    !===========================================================================
    !
    ! Lecture des options en entrée de l'application
    !

    ! Fichier de genotype
    !---------------------------------
    option_l  = "geno"
    option_s_opt = 'g'

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no genotype file defined *"
      stop
    end if

    genotypeFile = get_arg_string(option_l,option_s_opt)
    print *,"[genotype file=",trim(genotypeFile),"]"


    ! Nombre de caractère que prend l'identifiant dans le fichier de génotypes (10 par défaut)
    !-------------------------------------------------------------------------------------------
    option_l  = "idlength"
    option_s_opt = "i"
    defstr="10"
    sizeid = get_arg_string(option_l,option_s_opt,defstr)
    print *,"[sizeid=",trim(sizeid),"]"

    ! Le pattern des fichier mat_* (sortie du BLUPF90)
    ! L'application va chercher à lire les fichier <PATTERN><NUM_SNP>.dat

    option_l  = "matfile"
    option_s_opt = "m"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no pattern matrix file GBLUP defined *"
      stop
    end if

    patternMatFile = get_arg_string(option_l,option_s_opt)
    print *,"[pattern file=",trim(patternMatFile),"]"

    ! nombre de position testé
    option_l  = "npos"
    option_s_opt = "n"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no number of position to test defined *"
      stop
    end if

    nposstring = get_arg_string(option_l,option_s_opt)
    npos = get_arg_int(option_l,option_s_opt)
    print *,"[number of pos=",npos,"]"

    ! la variance residuelle
    option_l  = "vare"
    option_s_opt = "v"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no residual variance defined *"
      stop
    end if

    vare = get_arg_float(option_l,option_s_opt)
    print *,"[vare=",vare,"]"
    invvare = 1.d0 / vare

    option_l  = "nlevel"
    option_s_opt = "l"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no number of level defined *"
      stop
    end if

    nbe = get_arg_int(option_l,option_s_opt)
    print *,"[number of level (fixed effect)=",nbe,"]"

    option_l  = "animal"
    option_s_opt = "a"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no animal phenotyped file defined *"
      stop
    end if

    animalFile = get_arg_string(option_l,option_s_opt)

    print *,"[animal file=",trim(animalFile),"]"

    option_l  = "nsim"
    option_s_opt = "s"

    if ( .not. exist_arg(option_l,option_s_opt,.true.) ) then
      write (0,*) "* no number of simulation defined *"
      stop
    end if

    nsim = get_arg_int(option_l,option_s_opt)
    print *,"[number of simulation=",nsim,"]"

    print *
    print * ,"==========================================="
    print *

    if ( MAXPOSITION < npos ) then
     write (0,*) "MAXPOSITION to tested = ",MAXPOSITION
     stop
    end if

    call load_genotype(animalFile,genotypeFile,npos,id,genotype,nu)

    print *,"read ",nu," genotype."
    print *,"Genotype First ID [",id(1),"] =>",genotype(1,:10),"...",genotype(1,npos-10:)
    print *,"Genotype Last ID [",id(nu),"] =>",genotype(nu,:10),"...",genotype(nu,npos-10:)

    !Tableau pour retrouver les indices des animaux gere en interne
    allocate (revid(maxval(id)))
    revid=-1
    do i=1,nu
      revid(id(i))=i
    end do

    !====================================================================================
    !
    !
    ! Construction des Mqq
    !
    !====================================================================================

    ndim = nbe+1+nbe+nu+1 ! number of entries in file from blupf90 + 1 to manage last read

    allocate (A(npos,nbe),B(npos),C(npos,nbe),D(npos,nu))
    allocate (matrix_loaded(npos))
    allocate (M(npos,npos))

    matrix_loaded = .false.

    !$OMP PARALLEL DEFAULT(SHARED) PRIVATE(idmax,ivec,jvec,vvec,Eq1,Eq2,Fq1q2,qstr,q2,filemat)
    allocate (idmax(ndim),ivec(ndim),jvec(ndim),vvec(ndim),Eq1(nu),Eq2(nu))
    !$OMP DO
    do q1=1,npos
     ! On renumerote de 1 à 3 si l'allele 0 est representé dans la population le SNP q
     if ( count(genotype(:,q1) == 0 ) > 0 ) then
       where (genotype(:,q1) >=0 .and. genotype(:,q1) <=2 )
        Eq1 = genotype(:,q1)+1
       elsewhere
        Eq1 = 0.d0
       end where
     else
       Eq1 = genotype(:,q1)
       where ( Eq1 > 2 .or. Eq1 < 0 )
         Eq1 = 0.d0
       end where
     end if

      do q2=q1+1,npos
        !Si les fichiers ne sont pas lu, on garanti un access unique
        !$OMP CRITICAL
        if ( .not. matrix_loaded(q1)) then
           write (qstr,fmt="(I7)") q1
           filemat=trim(patternMatFile)//trim(adjustl(qstr))//".dat"
           call load_matrix(12345+omp_get_thread_num(),&
                            nu,&
                            nbe,&
                            filemat,&
                            idmax,&
                            ivec,&
                            jvec,&
                            vvec)
           call fill_matrix_snp(nu,nbe,idmax,jvec,vvec,A(q1,:),B(q1),C(q1,:),D(q1,:))
           matrix_loaded(q1) = .true.
        end if
        if ( .not. matrix_loaded(q2)) then
           write (qstr,fmt="(I7)") q2
           filemat=trim(patternMatFile)//trim(adjustl(qstr))//".dat"
           call load_matrix(12345+omp_get_thread_num(),&
                            nu,&
                            nbe,&
                            filemat,&
                            idmax,&
                            ivec,&
                            jvec,&
                            vvec)
           call fill_matrix_snp(nu,nbe,idmax,jvec,vvec,A(q2,:),B(q2),C(q2,:),D(q2,:))
           matrix_loaded(q2) = .true.
        end if
        !$OMP END CRITICAL

        ! On renumerote de 1 à 3 si l'allele 0 est representé dans la population le SNP q
        if ( count(genotype(:,q2) == 0 ) > 0 ) then
         where (genotype(:,q2) >=0 .and. genotype(:,q2) <=2 )
          Eq2 = genotype(:,q2)+1
         elsewhere
          Eq2 = 0.d0
         end where
        else
          Eq2 = genotype(:,q2)
          where ( Eq2 > 2 .or. Eq2 < 0)
            Eq2 = 0
          end where
        end if

        Fq1q2 = dot_product(Eq1,Eq2)

        !print *,Fq1q2
        M(q1,q2) = dot_product(C(q1,:),A(q2,:))*B(q2)
        !print *,M(q1,q2)
        M(q1,q2) = M(q1,q2) + (B(q1)*Fq1q2*B(q2))*invvare
        !print *,"==>",(B(q1)*Fq1q2*B(q2))*invvare,B(q1),B(q2),invvare,Fq1q2,"*"
        M(q1,q2) = M(q1,q2) + dot_product(D(q1,:),Eq2)*B(q2)*invvare
        M(q2,q1) = M(q1,q2)
     !   print *,q1,q2,M(q1,q2)
        !stop
      end do
      M(q1,q1) = B(q1)
     ! print *,q1,q1,M(q1,q1)
    end do
    !$OMP END DO
    deallocate (idmax,ivec,jvec,vvec,Eq1,Eq2)
    !$OMP END PARALLEL

    deallocate (matrix_loaded)
    deallocate (A,C,D)

    open(4567,file="M.txt",action="write",iostat=ios)
    if ( ios == 0 ) then
    print *,"======= write M.txt ====="
    else
      write (0,*) "Can not write M.txt"
    end if
    close (4567)


    !====================================================================================
    !
    ! Decomposition ACP de la matrice Mqq
    !
    !====================================================================================

    ! PARTI JM
    ! Etape 4 : Construction de la matrice de transformation

    allocate (SIGMA(npos),P(npos,npos))

    JOBZ='V'
    UPLO='U'
    allocate (M2(npos,npos),W(npos))
    M2 = M
    LWORK = 5*npos
    allocate (WORK(LWORK))
    call DSYEV( JOBZ, UPLO, npos, M2, npos, W, WORK, LWORK, INFO )
    print *,"ACP INFO=",info,WORK(1)
  !  print *,"Eigenvalues=",W
    print *,"size(Eigenvalues<0) =",count(W < 0)

    P = M2
    SIGMA = W

    deallocate (W,M2,WORK)

    where ( SIGMA < 0.d0 )
      SIGMA = 0.d0
    end where

    allocate (P2(npos,npos))
    ! Calcul de P2
    do i=1,npos
      P2(i,:) = P(i,:)*sqrt(SIGMA)
    end do

    ! Calcul de la diagonal de P3 = P * SIGMA * P'
    ! Diag P3 =>
    ! P3 (1,1) = p11*p11*SIGMA(1) + p12*p12*SIGMA(2) + ...

    allocate (P3diag(npos))

    do i=1,npos
       P3diag(i) = 0
       do j=1,npos
         P3diag(i) = P3diag(i) + P(i,j)*P(i,j)*SIGMA(j)
       end do
    end do

    deallocate (M,P,SIGMA)

    allocate (Psmin(nsim))
    allocate (Ts(npos),Pvalues(npos))

    ! Etape 5 : Simulation des seuils
    allocate (U(npos))

    open(4567,file="SIMUL.txt",action="write",iostat=ios)
    if ( ios == 0 ) then
     print *,"======= write SIMUL.txt ====="
    end if
    Pvalues=0.d0
    call init_random()
    TAIL_G01ECF = 'U'
    print *,"============= SIMUL ===================="
    do isim=1,nsim
      do s=1,npos
       u(s) = random_normal()
      end do
      !print *,"us:",u
      ! Muq est P...
      u=matmul(P2,u)
     ! print *,"as:",u
      Ts=0.d0
      do i=1,npos
        Ts(i)= u(i)*u(i) / P3diag(i)
        !print *, Ts(i),Mdiag(i)
        Pvalues(i) = G01ECF(TAIL_G01ECF, Ts(i) , 1.d0, IFAIL)
    !    print *,Ts(i),Pvalues(i),M(i,i)

      end do
      Psmin(isim)=minval(Pvalues)
   !   stop
       if ( ios == 0 ) write (4567,*) Psmin(isim)
    end do

    !trie à bulle...
    sort = .true.
    do while ( sort )
      sort = .false.
      do i=1,nsim-1
        if ( Psmin(i) > Psmin(i+1)) then
           vtemp = Psmin(i)
           Psmin(i) = Psmin(i+1)
           Psmin(i+1) = vtemp
           sort = .true.
        end if
      end do
    end do

    !print *,Psmin

    i = nint (( real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    print *,"1% => ",Psmin(i)

    i = nint (( 5 * real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    print *,"5% => ",Psmin(i)

    i = nint (( 10 * real(nsim) ) / 100.d0)
    if ( i == 0 ) i = 1
    print *,"10% => ",Psmin(i)

    deallocate (P2,P3diag)
    deallocate (Ts,Psmin,Pvalues)
    deallocate (U)
    deallocate (id,revid,genotype)


    contains

       subroutine load_genotype(animalFile,genotypeFile,npos,id,genotype,nu)
         character(len=len_long_cli)               ,intent(in) :: animalFile
         character(len=len_long_cli)               ,intent(in) :: genotypeFile
         integer                          ,intent(in) :: npos
         integer , dimension(:), pointer,intent(inout)    :: id
         integer , dimension(:,:), pointer,intent(inout)  :: genotype
         integer                          ,intent(inout)  :: nu

         !Local
         integer :: ios,s,i,unit_val=1234,izero
         integer :: nup = 0 ! number of animal to included in the analysis
         integer      ,dimension(:) ,allocatable   :: idfilter ! filtre des animaux a integrer (animaux phenotypé qu'on retrouve dans le fichier resultat de blupf90)
         integer      ,dimension(:) ,allocatable   :: revidfilter ! correspondance animaux -> id interne

         open(unit=unit_val,file=animalFile,iostat=ios,action="read")

         if ( ios /= 0 ) then
          write (0,*) "Can not open file ",trim(animalFile)
          stop
         end if

         nup = 0
         do while ( ios == 0 )
          nup=nup+1
          read (unit_val,'(a)',iostat=ios) bshort
         end do

         nup=nup-1
         rewind(unit_val)
         print *,"allocation size=",nup

         allocate (idfilter(nup))
         nup=0
         idfilter=0
         ios=0
         do while ( ios == 0 )
          read (unit_val,'(I'//trim(sizeid)//')',iostat=ios) idread

          if ( ios == 0 ) then
            nup=nup+1
            idfilter(nup) = idread
          end if
         end do

         close(unit_val)

         if ( nup <= 0 ) then
          write (0,*) "None element read in ",trim(animalFile)
          stop
         end if

         s = maxval(idfilter)
         print *,"allocate ",s," elements [revidfilter]"
         allocate (revidfilter(s))


         do i=1,nup
          revidfilter(idfilter(i))=i
         end do

         deallocate (idfilter)

        !Lecture des genotypes/haplotypes status
        open(unit=unit_val,file=genotypeFile,iostat=ios,action="read")

        if ( ios /= 0 ) then
         write (0,*) "Can not open file ",trim(genotypeFile)
         stop
        end if

        !
        !  LECTURE DES ID/GENOTYPES
        !
        allocate (id(nup))
        allocate (genotype(nup,npos))

        nu=0
        ios=0
        izero = ichar('0')
        genotype=-1
        id=-1
        rewind(unit_val)
        do while ( ios == 0 )
         idread = -1
         read (unit_val,'(I'//trim(sizeid)//',1x,(a'//trim(nposstring)//'))',iostat=ios) idread,bufsnp
         if ( ios == 0 .and. idread > 0 ) then
          if ( revidfilter(idread)>0) then
           nu = nu+1
           id(nu)=idread
           genotype(nu,:) = (/ (ichar(bufsnp(i:i))-izero,i=1,npos) /)
           revidfilter(idread) = 0
          end if
         end if
        end do

       close(unit_val)

       if ( nu <= 0 ) then
        write (0,*) "Can not read genotype file :[",trim(genotypeFile),"]"
        stop
       end if

       if ( nu /= nup ) then
        write (0,*) "bad correspondance between animal file and genotype file:"
        do i=1,nup
         if ( revidfilter(i)/=0) then
          write (0,*) "Can not finad animal ",revidfilter(i)," in genotype file."
         end if
        end do
        stop
       end if

       deallocate (revidfilter)

       end subroutine load_genotype


        ! Read file from blupf90
        ! Fist column 0 => fill A, 1=> fill B, 2 => fill C, 3=> fill D
        ! Second column => Alway 1 fro SNP model
        ! Third column  => level of effect (index)
        ! Forth column  => value
        subroutine load_matrix(idios,nu,nb,file_mat,idmat_vect,i_vect,j_vect,values_vect)
           integer            , intent(in) :: idios
           integer            , intent(in) :: nu ! number of animal
           integer            , intent(in) :: nb ! number of level for all fixed effect
           character(len=len_long_cli) , intent(in) :: file_mat
           integer            , intent(inout) :: idmat_vect(nb+1+nb+nu+1)
           integer            , intent(inout) :: i_vect(nb+1+nb+nu+1)
           integer            , intent(inout) :: j_vect(nb+1+nb+nu+1)
           real(kind=dp)      , intent(inout) :: values_vect(nb+1+nb+nu+1)

           integer :: ios,idx
           ! On met les valeurs dans des tableaux pour optimiser les transferts
           ! le dernier '1' permet de gerer la derniere lecture....


           open(idios,file=file_mat,iostat=ios,action="read")

           if ( ios /= 0 ) then
             write(0,*) "Can not open file [",trim(file_mat),"]"
             stop
           end if
           idx=0
           idmat_vect=-1
           do while (ios == 0 )
              idx=idx+1
              read (idios,*,iostat=ios) idmat_vect(idx), &
                                        i_vect(idx), &
                                        j_vect(idx), &
                                        values_vect(idx)
           end do

           close(idios)
           !print *,"[",trim(file_mat),"] ",idx-1," entries"


        end subroutine load_matrix

        subroutine fill_matrix_snp(nu,nb,idmat_vect,j_vect,values_vect,A,B,C,D)
           integer            , intent(in) :: nu ! number of animal
           integer            , intent(in) :: nb ! number of level for all fixed effect
           integer            , intent(in) :: idmat_vect(nb+1+nb+nu+1)
           integer            , intent(in) :: j_vect(nb+1+nb+nu+1)
           real(kind=dp)      , intent(in) :: values_vect(nb+1+nb+nu+1)
           real(kind=dp)      , intent(inout) , dimension(nb) :: A
           real(kind=dp)      , intent(inout)                 :: B
           real(kind=dp)      , intent(inout) , dimension(nb) :: C
           real(kind=dp)      , intent(inout) , dimension(nu) :: D

           integer :: ndim,i,centries

           ndim = nb+1+nb+nu

           do i=1,ndim
             select case (idmat_vect(i))
               case (0)
                A(j_vect(i)) = values_vect(i)
               case (1)
                B = values_vect(i)
               case (2)
                C(j_vect(i)) = values_vect(i)
               case (3)
                D(revid(j_vect(i))) = values_vect(i)
               case default

             end select
           end do

        end subroutine fill_matrix_snp

        subroutine fill_matrix_haplotype(nu,nb,nhap,idmat_vect,i_vect,j_vect,values_vect,A,B,C,D)
           integer            , intent(in) :: nu ! number of animal
           integer            , intent(in) :: nb ! number of level for all fixed effect
           integer            , intent(in) :: nhap ! number of level for the Haplotype effect (==1 for SNP)
           integer            , intent(in) :: idmat_vect(nb+1+nb+nu+1)
           integer            , intent(in) :: i_vect(nb+1+nb+nu+1)
           integer            , intent(in) :: j_vect(nb+1+nb+nu+1)
           real(kind=dp)      , intent(in) :: values_vect(nb+1+nb+nu+1)
           real(kind=dp)      , intent(inout) , dimension(nhap,nb)   :: A
           real(kind=dp)      , intent(inout) , dimension(nhap,nhap) :: B
           real(kind=dp)      , intent(inout) , dimension(nhap,nb)   :: C
           real(kind=dp)      , intent(inout) , dimension(nhap,nu)   :: D

        end subroutine fill_matrix_haplotype

        !!  DESCRIPTION
!!      Estimability with a decomposition of cholesky method
!!
!!      X'.X = L . L*
!!      where L is a lower triangular matrix with strictly positive diagonal entries, and L* denotes the conjugate transpose of L
!!
!!      output : incidenceDesc%vecsol  : array of boolean : true if estim , false otherwise
!!              incidenceDesc%nbnivest : number of effect to estimate according the decompostion
!!  INPUTS
!!    XX                         : X'.X
!!    sizeTriangComp             : size of the vector result
!!  OUTPUTS
!!    incidenceDesc%vecsol       : a logical array with the estimability of each level incidence matrix
!!    incidenceDesc%nbnivest     : the number of estimable level
!!    triangComp                 : X' V-1 X
!!
!!  NOTES
!!    triangComp is using in the resolution for the linear context
!!
!!  SOURCE
       subroutine estim_cholesky(sX,XX,triang,Lq)
         integer , intent(in)                             :: sX
         real (kind=dp) , dimension(sX,sX), intent(in)    :: XX
         real (kind=dp) , dimension(sX,sX), intent(out)   :: triang
         logical, dimension(sX) , intent(out)             :: Lq

         integer                                          :: i,j,k
         real (kind=dp)                                   :: threshold = 1d-2


         !  application de la decomposition de choleski pour determiner les contraintes
         !
         !  Dans le vecteur vecsol les effets a estimer sont indiques a TRUE et ceux
         !  qui ne sont pas estimables a FALSE
         !
         !  La matrice triang est la matrice triangulaire superieure M telle que
         !  M'M = X'X
         !
         Lq=.true.
         triang=0.d0
         do j=1,sX
           triang(j,j)=xx(j,j)
           do k=1,j-1
            triang(j,j)=triang(j,j)- triang(j,k)*triang(j,k)
           end do
           if (triang(j,j) > threshold ) then
             triang(j,j)=dsqrt(triang(j,j))
             do i=j+1,sX
               triang(i,j)=xx(i,j)
               do k=1,j-1
                triang(i,j)=triang(i,j)-triang(i,k)*triang(j,k)
               end do
               triang(i,j)=triang(i,j)/triang(j,j)
               triang(j,i)=triang(i,j)
             end do
            else
               Lq(j) = .false.
            end if
         end do

       end subroutine estim_cholesky

       subroutine init_random()
        integer :: n,clock,i
        integer, dimension(:), allocatable :: seed

        !initialisation pour les routines intrasec de fortran
        call random_seed(size=n)
        allocate(seed(n))
        call system_clock(count=clock)
        seed = clock + 37 * (/ (i - 1, i = 1, n) /)
        call random_seed(put = seed)
        deallocate(seed)

    end subroutine init_random

    subroutine rand_number(u)
        real , intent(out) :: u
        call random_number(u)
    end subroutine rand_number

     function normal(mean,sigma) !returns a normal distribution
        real(kind=dp)  , intent(in) :: mean,sigma
        real(kind=dp) normal,tmp
        integer flag
        real(kind=dp) fac,gsave,rsq,r1,r2,u
        save flag,gsave
        data flag /0/
        !$omp threadprivate (flag)
        if (flag.eq.0) then
        rsq=2.0
            do while(rsq.ge.1.0.or.rsq.eq.0.0) ! new from for do
                call random_number(u)
                r1=2.0*u-1.0
                call random_number(u)
                r2=2.0*u-1.0
                rsq=r1*r1+r2*r2
            enddo
            fac=sqrt(-2.0*log(rsq)/rsq)
            gsave=r1*fac
            tmp=r2*fac
            flag=1
        else
            tmp=gsave
            flag=0
        endif
        normal=tmp*sigma+mean
        return
    end function normal


FUNCTION random_normal() RESULT(fn_val)

! Adapted from the following Fortran 77 code
!      ALGORITHM 712, COLLECTED ALGORITHMS FROM ACM.
!      THIS WORK PUBLISHED IN TRANSACTIONS ON MATHEMATICAL SOFTWARE,
!      VOL. 18, NO. 4, DECEMBER, 1992, PP. 434-435.

!  The function random_normal() returns a normally distributed pseudo-random
!  number with zero mean and unit variance.

!  The algorithm uses the ratio of uniforms method of A.J. Kinderman
!  and J.F. Monahan augmented with quadratic bounding curves.

REAL :: fn_val

!     Local variables
REAL     :: s = 0.449871, t = -0.386595, a = 0.19600, b = 0.25472,    &
            r1 = 0.27597, r2 = 0.27846, u, v, x, y, q

!     Generate P = (u,v) uniform in rectangle enclosing acceptance region

DO
  CALL RANDOM_NUMBER(u)
  CALL RANDOM_NUMBER(v)
  v = 1.7156 * (v - half)

!     Evaluate the quadratic form
  x = u - s
  y = ABS(v) - t
  q = x**2 + y*(a*y - b*x)

!     Accept P if inside inner ellipse
  IF (q < r1) EXIT
!     Reject P if outside outer ellipse
  IF (q > r2) CYCLE
!     Reject P if outside acceptance region
  IF (v**2 < -4.0*LOG(u)*u**2) EXIT
END DO

!     Return ratio of P's coordinates as the normal deviate
fn_val = v/u
RETURN

END FUNCTION random_normal

   function G01ECF(TAIL, X, DF, IFAIL) result(res)
      integer , intent(out)        :: IFAIL
      real(kind=dp),intent(in)     :: X, DF
      character(len=1) ,intent(in) :: TAIL

      integer                      :: ind
      real                         :: Xreal,res
      real, external :: prob

      ind=int(DF)
      Xreal=X
      RES=prob(Xreal,ind)

      IF ( TAIL == 'U' ) RETURN
      IF ( TAIL == 'L') RES = 1-RES

   end function G01ECF

end program
