module cli
    implicit none

    ! Maximum length of name option
    integer , public , parameter :: len_long_cli=200

    private ::get_arg
    ! True if option are given in argument
    public :: exist_arg
    ! get integer value according an option
    public :: get_arg_int
    public :: get_arg_float
    public :: get_arg_string

    contains

      function get_arg(option_l,option_s,isval,val) result(find)
       character(len=len_long_cli) ,intent(in)          :: option_l
       character(len=1)        ,intent(in)          :: option_s
       logical                 ,intent(in)          :: isval
       character(len=len_long_cli) ,intent(out)         :: val

       character(len=len_long_cli) :: option
       integer :: i,nbargs
       logical :: find

       find = .false.
       val="<not-defined>"

       !idx is the index of the short option name
       nbargs = command_argument_count()

       do i=1,nbargs
         call get_command_argument(i,option)
         if ( (option == "--"//trim(option_l)) .or. (option == "-"//trim(option_s)) ) then
           find = .true.
           if (isval .and. i<nbargs) then
             call get_command_argument(i+1,val)
           else
             find = .false.
           end if
           return
         end if

       end do

      end function get_arg

      function exist_arg(option_l,option_s_opt,isval) result (find)
        character(len=len_long_cli) ,intent(in)          :: option_l
        character(len=1)        ,intent(in),optional :: option_s_opt
        logical                 ,intent(in)          :: isval

        logical :: find
        character(len=len_long_cli) :: val="",option_s

        find = .false.

        option_s = ""
        if ( present(option_s_opt)) option_s = option_s_opt

        find = get_arg(option_l,option_s,isval,val)


      end function exist_arg

      function get_arg_int(option_l,option_s,default) result (valint)
        character(len=len_long_cli) ,intent(in)          :: option_l
        character(len=1)        ,intent(in)              :: option_s
        integer                         ,optional        :: default

        logical :: find,isval
        character(len=len_long_cli) :: val=""
        integer :: valint

        valint = -1
        isval = .true.

        find = get_arg(option_l,option_s,isval,val)

        if ( .not. find ) then
          write (0,*) "Can not find option or integer value of [--"//trim(option_l)//",-"//option_s//"]"
          stop
        end if

        if ( .not. find .and. present(default)) then
         valint = default
        end if

        if (find) then
         read (val,fmt="(I7)") valint
        end if

      end function get_arg_int


      function get_arg_float(option_l,option_s_opt,default) result (valfloat)
        character(len=len_long_cli) ,intent(in)          :: option_l
        character(len=1)        ,intent(in),optional :: option_s_opt
        double precision                       ,optional :: default

        logical :: find,isval
        character(len=len_long_cli) :: val="",option_s
        double precision :: valfloat

        valfloat = -1.d0
        isval = .true.
        option_s = ""
        if ( present(option_s_opt)) option_s = option_s_opt

        find = get_arg(option_l,option_s,isval,val)

        if ( .not. find .and. present(default)) then
         valfloat = default
        end if

        if (find) then
         read (val,fmt="(F8.5)") valfloat
        end if

      end function get_arg_float

      function get_arg_string(option_l,option_s_opt,default) result (val)
        character(len=len_long_cli) ,intent(in)           :: option_l
        character(len=1)        ,intent(in),optional      :: option_s_opt
        character(len=len_long_cli)             ,optional :: default

        logical :: find,isval
        character(len=len_long_cli) :: option_s
        character(len=len_long_cli) :: val

        isval = .true.
        option_s = ""
        if ( present(option_s_opt)) option_s = option_s_opt

        find = get_arg(option_l,option_s,isval,val)

        if ( .not. find .and. present(default)) val = default

      end function get_arg_string



end module cli
